from __future__ import absolute_import

import os
from . import BaseProxy
from ..requests import FileRequest, munchify, POST
from ..helpers import for_all_methods, bind_proxy


@for_all_methods(bind_proxy)
class ModuleProxy(BaseProxy):

    def build_from_url(self, ownername, projectname, url, branch="master",
                       distgit=None):
        """
        Build a module from a URL pointing to a modulemd YAML file

        :param str ownername:
        :param str projectname:
        :param str url: URL pointing to a raw .yaml file
        :param str branch:
        :return: Munch
        """
        endpoint = "/module/build/{ownername}/{projectname}"
        params = {
            "ownername": ownername,
            "projectname": projectname,
        }
        data = {
            "scmurl": url,
            "branch": branch,
        }
        if distgit is not None:
            data["distgit"] = distgit
        response = self.request.send(
            endpoint=endpoint,
            method=POST,
            params=params,
            data=data,
            auth=self.auth,
        )
        return munchify(response)

    def build_from_file(self, ownername, projectname, path, distgit=None):
        """
        Build a module from a local modulemd YAML file

        :param str ownername:
        :param str projectname:
        :param str path:
        :return: Munch
        """
        endpoint = "/module/build/{ownername}/{projectname}"
        f = open(path, "rb")
        params = {
            "ownername": ownername,
            "projectname": projectname,
        }
        files = {
            "modulemd": (os.path.basename(f.name), f, "application/x-rpm")
        }
        data = None
        if distgit is not None:
            data = {"distgit": distgit}
        request = FileRequest(
            files=files,
            api_base_url=self.api_base_url,
            connection_attempts=self.config.get("connection_attempts", 1)
        )
        response = request.send(
            endpoint=endpoint,
            method=POST,
            params=params,
            data=data,
            auth=self.auth,
        )
        return munchify(response)



from requests import Response
from copr.test import mock
from copr.v3.requests import Request, munchify


class TestResponse(object):
    def test_munchify(self):
        response = mock.Mock(spec=Response)
        response.json.return_value = {"id": 1, "foo": "bar"}
        response.headers = {"Status": "200 OK"}

        entity = munchify(response)
        assert entity.id == 1
        assert entity.foo == "bar"
        assert entity.__response__ == response
        assert entity.__response__.headers["Status"] == "200 OK"
        assert entity.__response__.json()["foo"] == "bar"


class TestRequest(object):
    def test_endpoint_url(self):
        r1 = Request(api_base_url="http://copr/api_3")
        assert r1.endpoint_url("foo") == "http://copr/api_3/foo"

        # Leading and/or trailing slash should not be a problem
        r2 = Request(api_base_url="http://copr/api_3/")
        assert r2.endpoint_url("/foo/bar") == "http://copr/api_3/foo/bar"

    @mock.patch('requests.Session.request')
    def test_send(self, request):
        req1 = Request(api_base_url="http://copr/api_3")
        resp1 = req1.send(endpoint="foo")

        request.assert_called_once()
        args, kwargs = request.call_args
        assert kwargs["method"] == "GET"
        assert kwargs["url"] == "http://copr/api_3/foo"



from __future__ import absolute_import
from . import BaseProxy
from .build import BuildProxy
from ..requests import munchify, POST
from ..helpers import for_all_methods, bind_proxy


@for_all_methods(bind_proxy)
class PackageProxy(BaseProxy):

    def get(self, ownername, projectname, packagename,
            with_latest_build=False, with_latest_succeeded_build=False):
        """
        Return a package

        :param str ownername:
        :param str projectname:
        :param str packagename:
        :param bool with_latest_build: The result will contain "builds" dictionary with the latest
                                       submitted build of this particular package within the project
        :param bool with_latest_succeeded_build: The result will contain "builds" dictionary with
                                                 the latest successful build of this particular
                                                 package within the project.
        :return: Munch
        """
        endpoint = "/package"
        params = {
            "ownername": ownername,
            "projectname": projectname,
            "packagename": packagename,
            "with_latest_build": with_latest_build,
            "with_latest_succeeded_build": with_latest_succeeded_build,
        }
        response = self.request.send(endpoint=endpoint, params=params)
        return munchify(response)

    def get_list(self, ownername, projectname, pagination=None,
                 with_latest_build=False, with_latest_succeeded_build=False):
        """
        Return a list of packages

        :param str ownername:
        :param str projectname:
        :param pagination:
        :param bool with_latest_build: The result will contain "builds" dictionary with the latest
                                       submitted build of this particular package within the project
        :param bool with_latest_succeeded_build: The result will contain "builds" dictionary with
                                                 the latest successful build of this particular
                                                 package within the project.
        :return: Munch
        """
        endpoint = "/package/list"
        params = {
            "ownername": ownername,
            "projectname": projectname,
            "with_latest_build": with_latest_build,
            "with_latest_succeeded_build": with_latest_succeeded_build,
        }
        params.update(pagination or {})

        response = self.request.send(endpoint=endpoint, params=params)
        return munchify(response)

    def add(self, ownername, projectname, packagename, source_type, source_dict):
        """
        Add a package to a project

        :param str ownername:
        :param str projectname:
        :param str packagename:
        :param str source_type: http://python-copr.readthedocs.io/en/latest/client_v3/package_source_types.html
        :param dict source_dict: http://python-copr.readthedocs.io/en/latest/client_v3/package_source_types.html
        :return: Munch
        """
        endpoint = "/package/add/{ownername}/{projectname}/{package_name}/{source_type_text}"
        params = {
            "ownername": ownername,
            "projectname": projectname,
            "package_name": packagename,
            "source_type_text": source_type,
        }
        data = {
            "package_name": packagename,
        }
        data.update(source_dict)
        response = self.request.send(
            endpoint=endpoint,
            method=POST,
            params=params,
            data=data,
            auth=self.auth,
        )
        return munchify(response)

    def edit(self, ownername, projectname, packagename, source_type=None, source_dict=None):
        """
        Edit a package in a project

        :param str ownername:
        :param str projectname:
        :param str packagename:
        :param source_type: http://python-copr.readthedocs.io/en/latest/client_v3/package_source_types.html
        :param dict source_dict: http://python-copr.readthedocs.io/en/latest/client_v3/package_source_types.html
        :return: Munch
        """
        endpoint = "/package/edit/{ownername}/{projectname}/{package_name}/{source_type_text}"
        params = {
            "ownername": ownername,
            "projectname": projectname,
            "package_name": packagename,
            "source_type_text": source_type,
        }
        data = {
            "package_name": packagename,
        }
        data.update(source_dict or {})
        response = self.request.send(
            endpoint=endpoint,
            method=POST,
            params=params,
            data=data,
            auth=self.auth,
        )
        return munchify(response)

    def reset(self, ownername, projectname, packagename):
        """
        Reset a package configuration, meaning that previously selected
        ``source_type`` for the package and also all the source configuration
        previously defined by ``source_dict`` will be nulled.

        :param str ownername:
        :param str projectname:
        :param str packagename:
        :return: Munch
        """
        endpoint = "/package/reset"
        data = {
            "ownername": ownername,
            "projectname": projectname,
            "package_name": packagename,
        }
        response = self.request.send(
            endpoint=endpoint, data=data, method=POST, auth=self.auth)
        return munchify(response)

    def build(self, ownername, projectname, packagename, buildopts=None, project_dirname=None):
        """
        Create a build from a package configuration

        :param str ownername:
        :param str projectname:
        :param str packagename:
        :param buildopts: http://python-copr.readthedocs.io/en/latest/client_v3/build_options.html
        :param str project_dirname:
        :return: Munch
        """
        endpoint = "/package/build"
        data = {
            "ownername": ownername,
            "projectname": projectname,
            "package_name": packagename,
            "project_dirname": project_dirname,
        }
        build_proxy = BuildProxy(self.config)
        return build_proxy._create(endpoint, data, buildopts=buildopts)

    def delete(self, ownername, projectname, packagename):
        """
        Delete a package from a project

        :param str ownername:
        :param str projectname:
        :param str packagename:
        :return: Munch
        """
        endpoint = "/package/delete"
        data = {
            "ownername": ownername,
            "projectname": projectname,
            "package_name": packagename,
        }
        response = self.request.send(
            endpoint=endpoint, data=data, method=POST, auth=self.auth)
        return munchify(response)



"""
Authentication classes for usage within APIv3
"""

from copr.v3.exceptions import CoprAuthException
from copr.v3.auth.token import ApiToken
from copr.v3.auth.gssapi import Gssapi


def auth_from_config(config):
    """
    Decide what authentication method to use and return an appropriate instance
    """
    if config.get("token"):
        return ApiToken(config)

    if config.get("gssapi"):
        return Gssapi(config)

    msg = "GSSAPI disabled and login:token is invalid ({0}:{1})".format(
        config.get("login", "NOT_SET"),
        config.get("token", "NOT_SET"),
    )
    raise CoprAuthException(msg)



"""
Authentication via GSSAPI
"""

import requests

try:
    import requests_gssapi
except ImportError:
    requests_gssapi = None

from copr.v3.exceptions import CoprAuthException
from copr.v3.requests import munchify, handle_errors
from copr.v3.auth.base import BaseAuth


class Gssapi(BaseAuth):
    """
    Authentication via GSSAPI (i.e. Kerberos)
    """
    def __init__(self, *args, **kwargs):
        """
        Gssapi class stub for the systems where requests_gssapi is not
        installed (typically PyPI installations)
        """
        if not requests_gssapi:
            # Raise an exception if any dependency is not installed
            raise CoprAuthException(
                "The 'requests_gssapi' package is not installed. "
                "Please install it, or use the API token (config file)."
            )
        super(Gssapi, self).__init__(*args, **kwargs)

    def make_expensive(self):
        url = self.config["copr_url"] + "/api_3/gssapi_login/"
        auth = requests_gssapi.HTTPSPNEGOAuth(opportunistic_auth=True)
        try:
            response = requests.get(url, auth=auth)
        except requests_gssapi.exceptions.SPNEGOExchangeError as err:
            msg = "Can not get session for {0} cookie via GSSAPI: {1}".format(
                self.config["copr_url"], err)
            # TODO: change to `raise from` once we stop supporting rhel7
            raise CoprAuthException(msg)  # pylint: disable=raise-missing-from

        handle_errors(response)
        data = munchify(response)
        token = response.cookies.get("session")
        self.username = data.name
        self.cookies = {"session": token}



import time
import pytest
from copr.v3.auth import ApiToken, Gssapi, auth_from_config, CoprAuthException
from copr.v3.auth.base import BaseAuth
from copr.test import mock


class TestApiToken:
    @staticmethod
    def test_make_expensive():
        """
        Test that `ApiToken` load all necessary information from config
        """
        config = mock.MagicMock()
        auth = ApiToken(config)
        # Make sure all auth values are loaded from the config
        assert auth.auth
        assert auth.username


class TestGssApi:
    @staticmethod
    @mock.patch("requests.get")
    def test_make_expensive(mock_get):
        """
        Test that `Gssapi` knows what to do with information from Kerberos
        """
        config = mock.MagicMock()
        auth = Gssapi(config)
        assert not auth.username
        assert not auth.auth

        # Fake a response from Kerberos
        response = mock.MagicMock()
        response.json.return_value = {"id": 123, "name": "jdoe"}
        response.cookies = {"session": "foo-bar-hash"}
        mock_get.return_value = response

        # Make sure GSSAPI cookies are set
        auth.make_expensive()
        assert auth.username == "jdoe"
        assert auth.cookies == {"session": "foo-bar-hash"}


class TestBaseAuth:
    @staticmethod
    @mock.patch("copr.v3.auth.base.BaseAuth.expired")
    @mock.patch("copr.v3.auth.base.BaseAuth.make_expensive")
    def test_make_without_cache(make_expensive, _expired):
        """
        Test that auth classes remember and re-use the information from
        `make_expensive()`
        """
        auth = BaseAuth(config=None)
        auth.cache = mock.MagicMock()

        # Even though we call make() multiple times,
        # make_expensive() is called only once
        for _ in range(5):
            auth.make()
        assert make_expensive.call_count == 1

    @staticmethod
    @mock.patch("copr.v3.auth.base.BaseAuth.make_expensive")
    def test_make_reauth(make_expensive):
        """
        When reauth is requested, make sure we don't use any previous tokens
        """
        auth = BaseAuth(config=None)
        auth.cache = mock.MagicMock()
        for _ in range(5):
            auth.make(reauth=True)
        assert make_expensive.call_count == 5

    @staticmethod
    @mock.patch("copr.v3.auth.base.BaseAuth.make_expensive")
    def test_make_from_cache(make_expensive):
        """
        If there is a cached session cookie that is still valid, use it and
        don't make any expensive calls
        """
        auth = BaseAuth(config=None)
        auth.cache = FakeCache(None)
        for _ in range(5):
            auth.make()
        assert make_expensive.call_count == 0

    @staticmethod
    @mock.patch("copr.v3.auth.base.BaseAuth.make_expensive")
    def test_make_from_cache_expired(make_expensive):
        """
        If there is a cached session cookie but it is expired, just ignore it
        """
        auth = BaseAuth(config=None)
        auth.cache = FakeCacheExpired(None)
        for _ in range(5):
            auth.make()
        assert make_expensive.call_count == 1


class TestAuth:
    @staticmethod
    def test_auth_from_config():
        """
        Make sure we use the expected authentication method
        """
        # Use (login, token) authentication if there is enough information
        auth = auth_from_config({
            "copr_url": "http://copr",
            "login": "test",
            "token": "test",
            "username": "jdoe",
        })
        assert isinstance(auth, ApiToken)

        # Otherwise use GSSAPI (if gssapi is enabled, which is by default)
        auth = auth_from_config({
            "copr_url": "http://copr",
            "gssapi": True,
        })
        assert isinstance(auth, Gssapi)

        # There are no other authentication methods
        config = {
            "copr_url": "http://copr",
            "gssapi": False,
        }
        with pytest.raises(CoprAuthException):
            auth = auth_from_config(config)


class FakeCache(mock.MagicMock):
    # pylint: disable=too-many-ancestors
    def load_session(self):
        # pylint: disable=no-self-use
        return {
            "name": "jdoe",
            "session": "foo-bar-hash",
            "expiration": time.time() + 1
        }

class FakeCacheExpired(FakeCache):
    # pylint: disable=too-many-ancestors
    def load_session(self):
        return {
            "name": "jdoe",
            "session": "foo-bar-hash",
            "expiration": time.time() - 1
        }



"""
APIv3 /monitor Python client code
"""

from copr.v3 import proxies
from copr.v3.requests import munchify
from copr.v3.helpers import for_all_methods, bind_proxy


@for_all_methods(bind_proxy)
class MonitorProxy(proxies.BaseProxy):
    """
    Proxy to process /api_3/monitor requests.
    """

    def monitor(self, ownername, projectname, project_dirname=None,
                additional_fields=None):
        """
        Return a list of project packages, and corresponding info for the latest
        chroot builds.

        :param str ownername:
        :param str projectname:
        :param str project_dirname:
        :param list additional_fields: List of additional fields to return in
            the dictionary.  Possible values: ``url_build_log``,
            ``url_backend_log``, ``build_url``.  Note that additional fields
            may significantly prolong the server response time.
        :return: Munch a list of dictionaries,
            formatted like::

                {
                  "name": package_name,
                    "chroots": {
                      "fedora-rawhide-x86_64": {
                          "build_id": 843616,
                          "status": "succeeded",
                          ... fields ...,
                    }
                  },
                }
        """
        endpoint = "/monitor"
        params = {
            "ownername": ownername,
            "projectname": projectname,
            "project_dirname": project_dirname,
            "additional_fields[]": additional_fields,
        }
        response = self.request.send(endpoint=endpoint, params=params)
        return munchify(response)



"""
Authentication via (login, token)
"""

from copr.v3.auth.base import BaseAuth


class ApiToken(BaseAuth):
    """
    The standard authentication via `(login, token)` from `~/.config/copr`.
    """

    def __init__(self, *args, **kwargs):
        super(ApiToken, self).__init__(*args, **kwargs)
        # No need to lazy-login
        self.make()

    def make(self, reauth=False):
        """
        Override the higher abstraction method BaseAuth.make() because we don't
        need the complicated logic with API tokens at all (working with caches
        or sessions, etc.).  We also don't use make_expensive() at all.
        """
        self.auth = self.config["login"], self.config["token"]
        self.username = self.config.get("username")

    def make_expensive(self):
        """
        This is never called.  We have overridden the parent's make() method.
        """
        raise RuntimeError("ApiToken.make_expensive() should never be called")



import pytest
from munch import Munch
from copr.test import mock
from copr.v3.helpers import wait, succeeded, List
from copr.v3 import BuildProxy, CoprException


class TestHelpers(object):
    def test_succeeded(self):
        b1 = Munch(state="succeeded")
        b2 = Munch(state="succeeded")
        b3 = Munch(state="running")
        b4 = Munch(state="failed")

        assert succeeded(b1)
        assert succeeded([b1, b2])
        assert not succeeded(b3)
        assert not succeeded([b1, b2, b4])


class TestWait(object):
    @mock.patch("copr.v3.proxies.build.BuildProxy.get")
    def test_wait(self, mock_get):
        build = MunchMock(id=1, state="importing")

        mock_get.return_value = MunchMock(id=1, state="succeeded")
        assert wait(build)

        mock_get.return_value = MunchMock(id=1, state="unknown")
        with pytest.raises(CoprException) as ex:
            wait(build)
        assert "Unknown status" in str(ex)

    @mock.patch("copr.v3.proxies.build.BuildProxy.get")
    def test_wait_list(self, mock_get):
        builds = [MunchMock(id=1, state="succeeded"), MunchMock(id=2, state="failed")]
        mock_get.side_effect = lambda id: builds[id-1]
        assert wait(builds)

    @mock.patch("copr.v3.proxies.build.BuildProxy.get")
    def test_wait_custom_list(self, mock_get):
        builds = List([Munch(id=1, state="succeeded"), Munch(id=2, state="failed")],
                      proxy=BuildProxy({"copr_url": "http://copr", "login": "test", "token": "test"}))
        mock_get.side_effect = lambda id: builds[id-1]
        assert wait(builds)

    @mock.patch("time.time")
    @mock.patch("copr.v3.proxies.build.BuildProxy.get")
    def test_wait_timeout(self, mock_get, mock_time):
        build = MunchMock(id=1, state="importing")

        mock_get.return_value = MunchMock(id=1, state="running")
        mock_time.return_value = 0
        with pytest.raises(CoprException) as ex:
            wait(build, interval=0, timeout=-10)
        assert "Timeouted" in str(ex)

    @mock.patch("copr.v3.proxies.build.BuildProxy.get")
    def test_wait_callback(self, mock_get):
        build = MunchMock(id=1, state="importing")

        callback = mock.Mock()
        mock_get.return_value = MunchMock(id=1, state="failed")
        wait(build, interval=0, callback=callback)
        assert callback.called


class MunchMock(Munch):
    __proxy__ = BuildProxy({"copr_url": "http://copr", "login": "test", "token": "test"})



#!/usr/bin/python

import re
import os
import codecs

from setuptools import setup, find_packages

long_description = """Copr is designed to be a lightweight buildsystem that allows contributors
to create packages, put them in repositories, and make it easy for users
to install the packages onto their system. Within the Fedora Project it
is used to allow packagers to create third party repositories.

This part is a python client to the copr service."""

requires = [
    'filelock',
    'requests',
    'requests-toolbelt',
    'setuptools',
    'six',
    'munch',
]

__description__ = "Python client for copr service."
__author__ = "Valentin Gologuzov"
__author_email__ = "vgologuz@redhat.com"
__url__ = "https://github.com/fedora-copr/copr"


setup(
    name='copr',
    version="1.132",
    description=__description__,
    long_description=long_description,
    author=__author__,
    author_email=__author_email__,
    url=__url__,
    license='GPLv2+',
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        "Topic :: System :: Archiving :: Packaging",
        "Development Status :: 3 - Alpha",
    ],
    install_requires=requires,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
)



from __future__ import absolute_import

import os
import json
import time
import requests
from copr.v3.helpers import List
from munch import Munch
from requests_toolbelt.multipart.encoder import MultipartEncoder, MultipartEncoderMonitor
from .exceptions import CoprRequestException, CoprNoResultException, CoprTimeoutException, CoprAuthException


GET = "GET"
POST = "POST"
PUT = "PUT"


class Request(object):
    # This should be a replacement of the _fetch method from APIv1
    # We can have Request, FileRequest, AuthRequest/UnAuthRequest, ...

    def __init__(self, api_base_url=None, connection_attempts=1):
        """
        :param api_base_url:
        :param connection_attempts:

        @TODO maybe don't have both params and data, but rather only one variable
        @TODO and send it as data on POST and as params on GET
        """
        self.api_base_url = api_base_url
        self.connection_attempts = connection_attempts

    def endpoint_url(self, endpoint, params=None):
        params = params or {}
        endpoint = endpoint.strip("/").format(**params)
        return os.path.join(self.api_base_url, endpoint)

    def send(self, endpoint, method=GET, data=None, params=None, headers=None,
             auth=None):

        request_params = self._request_params(
            endpoint, method, data, params, headers, auth)

        response = self._send_request_repeatedly(request_params, auth)

        handle_errors(response)
        return response

    def _send_request_repeatedly(self, request_params, auth):
        """
        Repeat the request until it succeeds, or connection retry reaches its limit.
        """
        sleep = 5
        for i in range(1, self.connection_attempts + 1):
            try:
                response = requests.request(**request_params)
                if response.status_code == 401 and i < self.connection_attempts:
                    # try to authenticate again, don't sleep!
                    self._update_auth_params(request_params, auth, reauth=True)
                    continue
                # Return the response object (even for non-200 status codes!)
                return response
            except requests.exceptions.ConnectionError:
                if i < self.connection_attempts:
                    time.sleep(sleep)

        raise CoprRequestException("Unable to connect to {0}.".format(self.api_base_url))

    def _request_params(self, endpoint, method=GET, data=None, params=None,
                        headers=None, auth=None):
        params = {
            "url": self.endpoint_url(endpoint, params),
            "json": data,
            "method": method.upper(),
            "params": params,
            "headers": headers,
        }
        self._update_auth_params(params, auth)
        return params

    def _update_auth_params(self, request_params, auth, reauth=False):
        # pylint: disable=no-self-use
        if not auth:
            return

        auth.make(reauth)
        request_params.update({
            "auth": auth.auth,
            "cookies": auth.cookies,
        })


class FileRequest(Request):
    def __init__(self, files=None, progress_callback=None, **kwargs):
        super(FileRequest, self).__init__(**kwargs)
        self.files = files
        self.progress_callback = progress_callback

    def _request_params(self, *args, **kwargs):
        params = super(FileRequest, self)._request_params(*args, **kwargs)

        data = self.files or {}
        data["json"] = ("json", json.dumps(params["json"]), "application/json")

        callback = self.progress_callback or (lambda x: x)
        m = MultipartEncoder(data)
        params["json"] = None
        params["data"] = MultipartEncoderMonitor(m, callback)
        params["headers"] = {'Content-Type': params["data"].content_type}
        return params


def munchify(response):
    data = response.json()
    if "items" in data:
        return List(items=[Munch(obj) for obj in data["items"]],
                    meta=Munch(data["meta"]), response=response)
    return Munch(data, __response__=response)


def handle_errors(response):
    try:
        response_json = response.json()
        if "error" not in response_json:
            return

        if response.status_code == 403:
            raise CoprAuthException(response_json["error"], response=response)

        if response.status_code == 404:
            raise CoprNoResultException(response_json["error"], response=response)

        if response.status_code == 504:
            raise CoprTimeoutException(response_json["error"], response=response)

        raise CoprRequestException(response_json["error"], response=response)

    except ValueError:
        # When the request timeouted on the apache layer, we couldn't return a
        # nice JSON response and therefore its parsing fails.
        if response.status_code == 504:
            message = response.reason
            if response.request.method == "GET":
                message += "\nConsider using pagination for large queries."

            # We can't raise-from because of EPEL7
            # pylint: disable=raise-missing-from
            raise CoprTimeoutException(message, response=response)

        raise CoprRequestException("Response is not in JSON format, there is probably a bug in the API code.",
                                   response=response)
